// --------------------------------------------------------------------------
// External files inclusions.
// --------------------------------------------------------------------------
#include <pic.h>                                                // Use a PIC16F877A.
#include "io.h"
#include "time.h"



// --------------------------------------------------------------------------
// LCD command definitions, constants and timing settings.
// --------------------------------------------------------------------------
#define DDS_CMD_PHASE_ADJ_1     0x00        // Phase adjust register 1, followed by 2 bytes.
#define DDS_CMD_PHASE_ADJ_2     0x01        // Phase adjust register 2, followed by 2 bytes.
#define DDS_CMD_FREQ_WORD_1     0x02        // Frequency tuning word 1, followed by 6 bytes.
#define DDS_CMD_FREQ_WORD_2     0x03        // Frequency tuning word 2, followed by 6 bytes.
#define DDS_CMD_DELTE_FREQ_WORD 0x04        // Delta frequency word, followed by 6 bytes.
#define DDS_CMD_UPDATE_CLOCK    0x05        // Update clock, followed by 4 bytes.
#define DDS_CMD_RAMP_RATE_CLOCK 0x06        // Ramp rate clock, followed by 3 bytes.
#define DDS_CMD_CONTROL_REG     0x07        // Control register, followed by 4 bytes.
#define DDS_CMD_SHAPE_KEY_MULT  0x08        // Output shape key mult, followed by 2 bytes.
#define DDS_CMD_SHAPE_KEY_RR    0x0A        // Output shape key ramp rate, followed by 2 bytes.
#define DDS_CMD_CTRL_DAC        0x0B        // Control DAC, followed by 2 bytes.

const unsigned long int DDS_delay_reset = 3;                    // 23us

const long int DDS_clock = 300000000;                           // DDS system clock (after PLL multiplication).
const long int DDS_max_freq = 150000000;                        // Maximum output frequency (half of DDS_clock).
const long int DDS_min_freq = 0;                                // Minimum frequency is 0Hz, but output bandwidth may be limited.



// --------------------------------------------------------------------------
// This procedure will send 8 bit of data to the DDS, MSB first. The AD9852
// requires a clock period of at least 100ns: using a PIC16F877 with
// 18.432MHz clock each instruction takes about 217ns and we don't have to
// worry about communication timing.
// --------------------------------------------------------------------------
void DDS_send_data(unsigned char data)
{
    unsigned char i;                                            // Just a counter.
    
    for (i=0; i<8; i++)                                         // Loop for all 8 bit.
    {
        DDS_SDO = ((data & 0b10000000) == 0b10000000);          // Send leftmost bit of data.
        NOP();
        DDS_SCLK = 1;                                           // Rise SCLK.
        NOP();
        DDS_SCLK = 0;                                           // Drop SCLK.
        data = data << 1;                                       // Shift data left by 1 bit.
    }
}



// --------------------------------------------------------------------------
// This procedure will pulse the update clock to make the new change
// effective.
// --------------------------------------------------------------------------
void DDS_send_update(void)
{
    DDS_UPDCLK = 1;                                             // Set update clock.
    NOP();
    DDS_UPDCLK = 0;                                             // Clear update clock.
}



// --------------------------------------------------------------------------
// This procedure will reset the communication between the AD9852 and the
// PIC in case synchronization was lost. Registers of the DDS are not
// affected.
// --------------------------------------------------------------------------
void DDS_reset_comm(void)
{
    DDS_IORST = 1;                                              // Rise communication reset line.
    NOP();
    DDS_IORST = 0;                                              // Drop communication reset line.
}



// --------------------------------------------------------------------------
// This procedure will set the specified frequency (in Hz) into the DDS and
// make the change affective. DDS must be in "single tone mode".
// The frequency tuning word is 48 bit long and is calculated as follows:
//
// FTW = (freq * 2^48) / DDS_clock
//
// Which is equivalent to "freq" shifted left by 48 bit divided by
// "DDS_clock". "freq" is 32 bit long and when shifted left by 48 bit an 80
// bit variable is required. The denominator "DDS_clock" is also 32 bit long.
// This requires an integer division routine capable of dividing 80 bit by
// 32 bit providing 48 bit result (and 32 bit reminder, which is not used),
// which is not available in the C compiler and has been implemented in
// assembler. Because of the assembler code, all related variables must be
// global and in bank 1 (to save bank 0 space).
// --------------------------------------------------------------------------
bank1 unsigned char divisor[10];                                // The divisor (80 bit total) but also result of division (lowest 48 bit).
bank1 unsigned long int denominator, reminder;                  // Denominator and reminder of division.
bank1 unsigned char count, status_tmp;                          // Used to count bit during division and to temporary store the STATUS register.

void DDS_set_freq(long int freq)
{
    divisor[9] = (unsigned char) *(((char*)(&freq))+3);         // Put the desired frequency shifted left by 48 bit into divisor.
    divisor[8] = (unsigned char) *(((char*)(&freq))+2);         // Equivalent to "freq * 2^48".
    divisor[7] = (unsigned char) *(((char*)(&freq))+1);
    divisor[6] = (unsigned char) *(((char*)(&freq))+0);
    divisor[5] = 0;
    divisor[4] = 0;
    divisor[3] = 0;
    divisor[2] = 0;
    divisor[1] = 0;
    divisor[0] = 0;

    denominator = DDS_clock;                                    // Set denominator to perform "(freq * 2^48) / DDS_clock".

// This is the division routine: put the 80 bit divisor into "divisor" and the
// 32 bit denominator into "denominator. All variables must be in bank 1.
#asm
                swapf   0x03,w                                  ; Save STATUS into W.
                bsf     0x03,5                                  ; Select bank 1.
                bcf     0x03,6
                movwf   (_status_tmp & 0x7F)                    ; And save into _status_tmp

                clrf    ((_reminder & 0x7F)+0)                  ; Zero reminder before dividing (reminder = 0).
                clrf    ((_reminder & 0x7F)+1)
                clrf    ((_reminder & 0x7F)+2)
                clrf    ((_reminder & 0x7F)+3)

                movlw   80                                      ; Process 80 bit (count = 80)
                movwf   (_count & 0x7F)

loop:           bcf     0x03,0                                  ; Clear carry flag (insert 0).
                rlf     ((_divisor & 0x7F)+0),f                 ; Shift everything left by 1 bit.
                rlf     ((_divisor & 0x7F)+1),f
                rlf     ((_divisor & 0x7F)+2),f
                rlf     ((_divisor & 0x7F)+3),f
                rlf     ((_divisor & 0x7F)+4),f
                rlf     ((_divisor & 0x7F)+5),f
                rlf     ((_divisor & 0x7F)+6),f
                rlf     ((_divisor & 0x7F)+7),f
                rlf     ((_divisor & 0x7F)+8),f
                rlf     ((_divisor & 0x7F)+9),f                 ; Put lefmost bit of divisor into reminder.
                rlf     ((_reminder & 0x7F)+0),f
                rlf     ((_reminder & 0x7F)+1),f
                rlf     ((_reminder & 0x7F)+2),f
                rlf     ((_reminder & 0x7F)+3),f

                movf    ((_denominator & 0x7F)+3),w             ; Compare reminder (32 bit) with denominator (32 bit).
                subwf   ((_reminder & 0x7F)+3),w
                btfss   0x03,2                                  ; Check zero flag.
                goto    skip_lower_cmp
                movf    ((_denominator & 0x7F)+2),w
                subwf   ((_reminder & 0x7F)+2),w
                btfss   0x03,2                                  ; Check zero flag.
                goto    skip_lower_cmp
                movf    ((_denominator & 0x7F)+1),w
                subwf   ((_reminder & 0x7F)+1),w
                btfss   0x03,2                                  ; Check zero flag.
                goto    skip_lower_cmp
                movf    ((_denominator & 0x7F)+0),w
                subwf   ((_reminder & 0x7F)+0),w
skip_lower_cmp: btfss   0x03,0                                  ; Check carry flag.
                goto    not_yet                                 ; Denominator is greater than reminder, add a zero to the result (already done while shifting).
                
                bsf     ((_divisor & 0x7F)+0),0                 ; Denominator is less than or equal to reminder: add 1 to result.

                movf    ((_denominator & 0x7F)+0),w             ; Perform "reminder = reminder - denominator"
                subwf   ((_reminder & 0x7F)+0),f
                movf    ((_denominator & 0x7F)+1),w
                btfss   0x03,0                                  ; Check carry flag.
                incfsz  ((_denominator & 0x7F)+1),w
                subwf   ((_reminder & 0x7F)+1),f
                movf    ((_denominator & 0x7F)+2),w
                btfss   0x03,0                                  ; Check carry flag.
                incfsz  ((_denominator & 0x7F)+2),w
                subwf   ((_reminder & 0x7F)+2),f
                movf    ((_denominator & 0x7F)+3),w
                btfss   0x03,0                                  ; Check carry flag.
                incf    ((_denominator & 0x7F)+3),w
                subwf   ((_reminder & 0x7F)+3),f

not_yet:        decfsz  (_count & 0x7F)                         ; Loop for every bit.
                goto    loop

                swapf   (_status_tmp & 0x7F),w                  ; Restore STATUS register.
                movwf   0x03
#endasm
// The division is done: result is in "divisor" (48 lowest bit) and the
// reminder (which is not used here) is in "reminder".

    DDS_reset_comm();                                           // Reset communication interface.
    DDS_send_data(DDS_CMD_FREQ_WORD_1);                         // Set frequency tining word 1 (followed by 6 bytes).
    DDS_send_data(divisor[5]);                                  // Transmit 48 bit of data.
    DDS_send_data(divisor[4]);
    DDS_send_data(divisor[3]);
    DDS_send_data(divisor[2]);
    DDS_send_data(divisor[1]);
    DDS_send_data(divisor[0]);
    DDS_send_update();                                          // Send update clock to make the frequency effective.
}



// --------------------------------------------------------------------------
// This procedure will initialize the AD9852 DDS chip. First a master reset
// is performed, followed by a communication reset. The the DDS is
// configured as follows:
//
//  - Cosine DAC is powered.
//  - Control DAC and comparators are switched off.
//  - PLL is activated and set to 10 (for generating 300MHz from a 30MHz
//    oscillator.
//  - The update clock is changed from internal to external to drive updates
//    directly from the PIC. In order to do this the direction of the update
//  - Operating mode is set to "single tone" and modulations are disabled.
//  - The serial port is configured for using two separate SDI/SDO lines.
//
// The CS line is held low all the time.
// --------------------------------------------------------------------------
void DDS_init(void)
{
    DDS_RESET = 1;                                              // Reset DDS.
    delay(DDS_delay_reset);                                     // Reset pulse should last for at least 10 system clocks (33.3ns at 300MHz).
    DDS_RESET = 0;
    NOP();

    DDS_CS = 0;                                                 // Make sure CS is cleared.

    DDS_send_data(DDS_CMD_CONTROL_REG);                         // Set control register, followed by 4 bytes.
    DDS_send_data(0b00010100);
//                  
//                  +-- CR24 Digital power down (1 = power down). Digital section is used.
//                  +--- CR25 Full DAC (cosine DAC, control DAC and reference) power down
//                            (1 = power down). Cosine DAC is used.
//                  +---- CR26 Control DAC power down (1 = power down). Control DAC is not used.
//                  +----- CR27 Set always to 0.
//                  +------ CR28 Comparator power down (1 = power down). Comparator is not used.
//                  +------- CR29 Open.
//                  +-------- CR30 Open.
//                  +--------- CR31 Open.
    DDS_send_data(0b01001010);
//                  
//                  +++++-- CR16-CR20 PLL multiplication factor (4..20). 10 is used to get 300MHz out
//                               of 30MHz.
//                  +------- CR21 PLL bypass (0 = PLL enabled). PLL is used.
//                  +-------- CR22 PLL range: set to 1 for frequencies higher than 200MHz. 300MHz are used.
//                  +--------- CR23 Reserved: set to 0.
    DDS_send_data(0b00000000);
//                  
//                  +-- CR08 Internal update active (1 = internal, 0 = external). This pin changes
//                           the direction of the line. Set to 0 to set update clock from PIC to DDS.
//                  +++--- CR09-CR11 Operating mode (0 = single tone, 1 = FSK, 2 = ramped FSK,
//                              3 = Chirp, 4 = BPSK). Single tone is used.
//                  +------ CR12 Don't care. Set to 0.
//                  +------- CR13 Triangle function (1 = active). Not used.
//                  +-------- CR14 Clear accumulators (1 = clear). Not used.
//                  +--------- CR15 Clear accumulator 1 (1 = clear). Not used.
    DDS_send_data(0b00000001);
//                  
//                  +-- CR00 Serial port SDO active (0 = SDIO (default), 1 = SDI + SDO). Separate
//                           SDI and SDO lines are used.
//                  +--- CR01 Serial port MSB/LSB first (0 = MSB first (default), 1 = LSB first).
//                            MSB first is used.
//                  +---- CR02 Reserved. Must be set to 0.
//                  +----- CR03 Reserved. Must be set to 0.
//                  +------ CR04 Internal external output shaping keying (0 = internal). Internal is used.
//                  +------- CR05 Output shaped keying enable (1 = enable). Not used to get full amplitude.
//                  +-------- CR06 Inverse sinc filter bypass (1 = bypass). Sinc filter is used to make
//                                 output amplitude independent from frequency. Disable to save power.
//                  +--------- CR07 Reserved. Must be set to 0.

    DDS_UPDCLK_TRIS = 0;                                        // UPDCLK is now output.
}


