// --------------------------------------------------------------------------
// External files inclusions.
// --------------------------------------------------------------------------
#include <pic.h>                                                // Use a PIC16F877A.
#include "io.h"
#include "time.h"
#include "lcd.h"



// --------------------------------------------------------------------------
// LCD command definitions and timing settings.
// --------------------------------------------------------------------------
#define LCD_cmd_reset                   0x00
#define LCD_cmd_cls                     0x01
#define LCD_cmd_home                    0x02

#define LCD_cmd_entrymode               0x04
#define LCD_cmd_entrymode_inc           0x02                    // set: increment, clear: decrement.
#define LCD_cmd_entrymode_shift         0x01                    // set: display shift, clear: no shift.

#define LCD_cmd_display                 0x08
#define LCD_cmd_display_on              0x04                    // set: display on, clear: display off
#define LCD_cmd_display_cursor          0x02                    // set: cursor on, clear: cursor off
#define LCD_cmd_display_blink           0x01                    // set: blink, clear: don't blink

#define LCD_cmd_shift                   0x10
#define LCD_cmd_shift_disp              0x08                    // set: display, clear: cursor
#define LCD_cmd_shift_right             0x04                    // set: right, clear: left

#define LCD_cmd_function                0x20
#define LCD_cmd_function_8bit           0x10                    // set: 8bit, clear: 4bit
#define LCD_cmd_function_2lines         0x08                    // set: 2 lines, clear: 1 line
#define LCD_cmd_function_5x10           0x04                    // set: font 5x10, clear: font 5x7

#define LCD_cmd_setddram                0x80
#define LCD_cmd_setddram_line1          0x00
#define LCD_cmd_setddram_line2          0x40
#define LCD_cmd_setddram_line3          0x10
#define LCD_cmd_setddram_line4          0x50

const unsigned long int LCD_delay_cmd = 10;                     // 61us
const unsigned long int LCD_delay_8bit = 27;                    // 153us
const unsigned long int LCD_delay_long = 380;                   // 2.0ms
const unsigned long int LCD_delay_init = 15500;                 // 80ms



// --------------------------------------------------------------------------
// This procedure waits a short amount of time and is used for
// synchoronization of the LCD communication. The minimum time to wait is
// 4us. This procedure will delay about 5.2us at 18.432MHz.
// --------------------------------------------------------------------------
void LCD_delay_pulse(void)
{
    unsigned char i;
    for (i=0; i<3; i++);
}



// --------------------------------------------------------------------------
// This procedure sends the specified 8bit data to the LCD display pretending
// the interface is 8bit wide (even if in this circuit a 4bit interface is
// used). This procedure is necessary (and only used) for initializing the
// LCD. The interface is used in the write only mode (no data is read from
// the LCD) for saving one I/O line.
// --------------------------------------------------------------------------
void LCD_write_command_8bit(unsigned char data)
{
    LCD_RS = 0;
    NOP();
    LCD_EN = 0;
    NOP();
    LCD_DATA = (LCD_DATA & 0xF0) | (data >> 4);                 // Only the 4 MSB bits are transmitted.
    LCD_delay_pulse();
    LCD_EN = 1;
    LCD_delay_pulse();
    LCD_EN = 0;
}



// --------------------------------------------------------------------------
// This procedure sends the specified 8bit command to the LCD display over 
// the 4bit interface. The interface is used in the write only mode (no data
// is read from the LCD) for saving one I/O line.
// --------------------------------------------------------------------------
void LCD_write_command(unsigned char data)
{
    LCD_RS = 0;
    NOP();
    LCD_EN = 0;
    NOP();
    LCD_DATA = (LCD_DATA & 0xF0) | (data >> 4);                 // Send most significant nibble first.
    NOP();
    LCD_EN = 1;
    LCD_delay_pulse();
    LCD_EN = 0;
    LCD_delay_pulse();
    LCD_DATA = (LCD_DATA & 0xF0) | (data & 0x0F);               // Then send least significant nibble.
    NOP();
    LCD_EN = 1;
    LCD_delay_pulse();
    LCD_EN = 0;
    delay(LCD_delay_cmd);                                       // Wait enough time to make sure the LCD handeled the data.
}



// --------------------------------------------------------------------------
// This procedure sends the specified 8bit data to the LCD display over 
// the 4bit interface. The interface is used in the write only mode (no data
// is read from the LCD) for saving one I/O line.
// --------------------------------------------------------------------------
void LCD_write_data(unsigned char data)
{
    LCD_RS = 1;
    NOP();
    LCD_EN = 0;
    NOP();
    LCD_DATA = (LCD_DATA & 0xF0) | (data >> 4);                 // Send most significant nibble first.
    NOP();
    LCD_EN = 1;
    LCD_delay_pulse();
    LCD_EN = 0;
    LCD_delay_pulse();
    LCD_DATA = (LCD_DATA & 0xF0) | (data & 0x0F);               // Then send least significant nibble.
    NOP();
    LCD_EN = 1;
    LCD_delay_pulse();
    LCD_EN = 0;
    delay(LCD_delay_cmd);                                       // Wait enough time to make sure the LCD handeled the data.
}



// --------------------------------------------------------------------------
// This procedure switches the LCD display on and turns the cursor off.
// --------------------------------------------------------------------------
void LCD_on_nocursor(void)
{
    LCD_write_command(LCD_cmd_display | LCD_cmd_display_on);    // Switch LCD display on.
}



// --------------------------------------------------------------------------
// This procedure switches the LCD display on and makes the cursor block
// and blinking.
// --------------------------------------------------------------------------
void LCD_on_cursor(void)
{
    LCD_write_command(LCD_cmd_display | LCD_cmd_display_on | LCD_cmd_display_blink);    // This will turn the block cursor on.
}



// --------------------------------------------------------------------------
// This procedure switches the LCD display off.
// --------------------------------------------------------------------------
void LCD_off(void)
{
    LCD_write_command(LCD_cmd_display);                         // Display off, cursor off, blink off.
}



// --------------------------------------------------------------------------
// This procedure clears the LCD display. It's a long command and therefore
// an extra delay is added.
// --------------------------------------------------------------------------
void LCD_clear(void)
{
    LCD_write_command(LCD_cmd_cls);                             // Clears LCD display.
    delay(LCD_delay_long);                                      // This command takes at least 1.52ms to execute.
}



// --------------------------------------------------------------------------
// This procedure returns the cursor of the LCD display to it's home
// position (0;0). It's a long command and therefore an extra delay is added.
// --------------------------------------------------------------------------
void LCD_home(void)
{
    LCD_write_command(LCD_cmd_home);                            // Put cursor to 0,0 position.
    delay(LCD_delay_long);                                      // This command takes at least 1.52ms to execute.
}



// --------------------------------------------------------------------------
// This procedure put the cursor of the LCD display at the beginning of line
// number 1 (the first one).
// --------------------------------------------------------------------------
void LCD_line_1(void)
{
    LCD_write_command(LCD_cmd_setddram | LCD_cmd_setddram_line1);   // Select LCD line 1.
    delay(LCD_delay_long);}



// --------------------------------------------------------------------------
// This procedure put the cursor of the LCD display at the beginning of line
// number 2.
// --------------------------------------------------------------------------
void LCD_line_2(void)
{
    LCD_write_command(LCD_cmd_setddram | LCD_cmd_setddram_line2);   // Select LCD line 2.
    delay(LCD_delay_long);}



// --------------------------------------------------------------------------
// This procedure shifts the cursor of the LCD display left by one place.
// --------------------------------------------------------------------------
void LCD_shift_cursor_left(void)
{
    LCD_write_command(LCD_cmd_shift);
}



// --------------------------------------------------------------------------
// This procedure shifts the cursor of the LCD display right by one place.
// --------------------------------------------------------------------------
void LCD_shift_cursor_right(void)
{
    LCD_write_command(LCD_cmd_shift | LCD_cmd_shift_right);
}



// --------------------------------------------------------------------------
// This procedure writes the specified string at the cursor position on the
// LCD display.
// --------------------------------------------------------------------------
void LCD_write_string(const unsigned char * string)
{
    unsigned char char_pointer;                                 // Pointer at the current character.
    
    char_pointer = 0;                                           // Start with first character.
    while (string[char_pointer] != 0)                           // Parse the string untill a null is found.
    {
        LCD_write_data(string[char_pointer]);                   // Write character.
        char_pointer++;
    }
}



// --------------------------------------------------------------------------
// This procedure initializes the LCD display. It must be executed once at
// powerup before using the display and takes about 0.1s to execute.
// --------------------------------------------------------------------------
void LCD_init(void)
{
    delay(LCD_delay_init);                                              // First wait 80ms
    LCD_write_command_8bit(LCD_cmd_function | LCD_cmd_function_8bit);   // Pretend interface is 8bit.
    delay(LCD_delay_init);                                              // Wait 80ms
    LCD_write_command_8bit(LCD_cmd_function | LCD_cmd_function_8bit);   // Pretend interface is 8bit again.
    delay(LCD_delay_8bit);                                              // Wait 150us
    LCD_write_command_8bit(LCD_cmd_function | LCD_cmd_function_8bit);   // Pretend interface is 8bit again.
    delay(LCD_delay_8bit);                                              // Wait 150us
    LCD_write_command_8bit(LCD_cmd_function);                           // Now interface is 8bit.
    delay(LCD_delay_8bit);                                              // Wait 150us
    LCD_write_command(LCD_cmd_function | LCD_cmd_function_2lines);      // 4 bit mode, 2 lines.
    LCD_write_command(LCD_cmd_reset);
    LCD_clear();                                                        // Display off, cursor off, blink off, increment off.
    LCD_on_nocursor();
    LCD_write_command(LCD_cmd_entrymode | LCD_cmd_entrymode_inc);       // Set auto increment, no shift.
}


