/*===========================================================================



AD9852 DDS generator
====================



Overview:
=========

CPU         :   PIC16F877A
Clock       :   XTAL 18.432MHz HS oscillator
Compiler    :   Hitech PICC
Version     :   1.00
Date        :   22.02.2006
Author      :   Iacopo Giangrandi HB9DUL



I/O connections:
================

See io.h for description.



Description:
============

This program is intended to control an AD9852 DDS as a VFO frequency
generator. The program simulates two VFOs that can be switched with a
pushbutton. The frequency can be adjusted with a rotary encoder, typed on
a numeric keypad or set via the RS232 port. It uses a 2 lines 16
characters alphanumeric LCD display.


Encoder / buttons:
------------------

The rotary encoder will increase or decrease the frequency by a given step
that can also be adjusted with the same encoder. Switching form frequency
adjust mode to step adjust mode is done by pressing the encoder knob. Each
VFO has its own step.

A button toggles between VFO A and B immediately changing the output
frequency. Another button will save the current settings into the EEPROM
memory and this will be the default parameters next time the DDS is
powered on. The frequencies of both VFOs, the frequency steps of both VFOs
and the active VFO are saved.


Keypad
------

The frequency can also be entered directly with a 12 key numeric keypad.
The "#" button acts as an "enter" button. When the ninth digit is entered
the "enter" is automatic, since the DDS cannot go over 150000000 Hz.
The "*" button acts as "MHz" button or as "esc" button. If a maximum of
3 digits have been entered, pressing "*" will act as a decimal point and the
frequency will be considered in MHz instead of Hz. Pressing "*" again or
after the third digit will cancel the operation. Pressing any other button
will cancle the operation as well.

Examples:

"1" "4" "0" "0" "0" "0" "0" "0" "#"                      14 000 000 Hz
"1" "4" "*" "#"                                          14 000 000 Hz
"1" "4" "*" "1" "#"                                      14 100 000 Hz
"1" "0" "0" "#"                                                 100 Hz
"0" "0" "0" "0" "0" "0" "1" "0" "0"                             100 Hz
"*" "0" "0" "0" "1" "#"                                         100 Hz
"1" "4" "5" "6" "0" "0" "0" "0" "0"                     145 600 000 Hz
"1" "4" "5" "6" "*"                                          no change
"1" "*" "*"                                                  no change


RS232 serial communication:
---------------------------

The frequency can also be entered via the RS232 port (9600N81) with the
following protocol:

FA14100000 [cr]     Set frequency of VFO A to 14 100 000 Hz
FB100 [cr]          Set frequency of VFO B to 100 Hz

SA [cr]             Set VFO A as active VFO
SB [cr]             Set VFO B as active VFO

? [cr]              Will return the program version.

Characters are not echoed back as typed, but when a [cr] is received
the whle command buffer is returned followed by [cr] [lf], and then,
ff the command is successfull the DDS controller will answer:

"Ok." [cr] [lf]

Otherwise the answer will be:

"Error: xxxxxx" [cr] [lf]

Where "xxxxxx" is a brief description of the error.

The serial commands are not case sensitive. Send a [cr] to reset the RS232
command handler before sending commands in order to make sure the buffer is
empty.

The RS232 handler is not interrupt driven: characters are received only when
the PIC is idle. Make sure the encoder or the keypad are not used when
commands are sent. While entering a frequency with the keypad the RS232
connection is never checked.



============================================================================= */




// --------------------------------------------------------------------------
// External files inclusions.
// --------------------------------------------------------------------------
#include <pic.h>                                                // Use a PIC16F877A.
#include "io.h"                                                 // I/O lines definitions.
#include "time.h"                                               // Delay procedure.
#include "lcd.h"                                                // LCD display handling routines.
#include "eeprom.h"                                             // EEPROM handling routines.
#include "dds.h"                                                // DDS handling routines.
#include "rs232.h"                                              // RS232 handling routines.



// --------------------------------------------------------------------------
// Configuration bits definitions.
// --------------------------------------------------------------------------
__CONFIG(HS & WDTDIS & PWRTEN & BOREN & LVPDIS & UNPROTECT & DEBUGDIS);
// Use external XTAL 18.432 MHz HS oscillator, do not use watchdog, enable power
// up timer, enable brown out detection, disable low voltage programming, do not
// use data protection and disable debugging.



// --------------------------------------------------------------------------
// VFO definitions, constants and global variables.
// --------------------------------------------------------------------------
#define WELCOME_STRING_1    "DDS VFO - AD9852"
#define WELCOME_STRING_2    "v1.00  by HB9DUL"

unsigned char VFO_active;                                       // VFO_A or VFO_B.
#define VFO_A   0x00
#define VFO_B   0x01

long int VFO_A_freq, VFO_B_freq;                                // Frequencies of both VFOs.
signed char VFO_A_step, VFO_B_step;                             // Setp indexes of both VFOs (for VFO_steps).
#define VFO_num_steps   24                                      // Number of different frequency steps available.
const long int VFO_steps[VFO_num_steps] =                       // Frequency step list, in Hz.

         {        1,       2,       5,
                 10,      20,      50,
                100,     200,     500,
               1000,    2000,    5000,    9000,
              10000,   12500,   25000,   50000,
             100000,  250000,  500000,
            1000000, 2000000, 5000000,
           10000000};
           
unsigned char VFO_edit_mode;                                    // VFO_edit_freq or VFO_edit_step.
#define VFO_edit_freq   0
#define VFO_edit_step   1



// --------------------------------------------------------------------------
// This procedure converts the specified frequency into an 11 character ASCII
// string and prints it on the LCD screen. Leading zeros are removed and
// digits are grouped by 3 by insetring spaces. If the frequency is negative
// it's absolute value is shown.
// --------------------------------------------------------------------------
void write_freq(long int freq)
{
    unsigned char ascii_freq[12];                               // Eleven characters string used to display the frequency in formatted ASCII decimal format.
    signed char i;                                              // Just a counter.

    if (freq < 0)                                               // Check for negative frequency.
        freq = -freq;                                           // Take absolute value.

    for (i=10; i>=0; i--)                                       // Convert into an ASCII string.
    {
        if ((i == 3) || (i == 7))                               // Insert a space every 3 digits.
            ascii_freq[i] = ' ';
        else
        {
            ascii_freq[i] = (freq % 10) + '0';                  // Get ASCII digit.
            freq /= 10;
        }
    }

    for (i=0; i<10; i++)                                        // Remove leading zeros.
        if ((i != 3) && (i != 7))                               // Skip spaces every 3 digits.
            if (ascii_freq[i] == '0')                           // Replace 0 with spaces.
                ascii_freq[i] = ' ';
            else
                break;                                          // Until the first non-zero digit is found.
    ascii_freq[11] = 0x00;                                      // Null terminate the string.

    LCD_write_string(ascii_freq);                               // Write result on LCD screen.
}



// --------------------------------------------------------------------------
// This procedure displays the specified VFO (VFO_A or VFO_B) information on
// the corresponding line of the LCD, including the vfo name ("A" or "B"), if
// it's active (by adding a "*"), its frequency with unit "Hz".
// --------------------------------------------------------------------------
void display_vfo(unsigned char vfo)
{
    if (vfo == VFO_A)                                           // Select correct line and display VFO name.
    {
        LCD_line_1();
        LCD_write_data('A');
    }
    else
    {
        LCD_line_2();
        LCD_write_data('B');
    }

    if (vfo == VFO_active)                                      // Mark active VFO with a "*" if it's the case.
        LCD_write_data('*');
    else
        LCD_write_data(' ');

    if (vfo == VFO_A)                                           // Display VFO frequency.
        write_freq(VFO_A_freq);
    else
        write_freq(VFO_B_freq);

    LCD_write_string(" Hz");                                    // Display unit.
}



// --------------------------------------------------------------------------
// This procedure displays the specified VFO (VFO_A or VFO_B) frequency step
// on the corresponding line of the LCD.
// --------------------------------------------------------------------------
void display_step(unsigned char vfo)
{
    if (vfo == VFO_A)                                           // Select correct line.
        LCD_line_1();
    else
        LCD_line_2();

    LCD_write_string("Step ");                                  // Tell about step edit mode.

    if (vfo == VFO_A)                                           // Display correct step frequency.
        write_freq(VFO_steps[VFO_A_step]);
    else
        write_freq(VFO_steps[VFO_B_step]);
}



// --------------------------------------------------------------------------
// This function converts the specified string into an integer number.
// No check is done: the input string must be composed by digits ('0'..'9')
// only. No negative nor real numbers are supported. Number must fit into
// a long int variable.
// --------------------------------------------------------------------------
long int str_2_long_int(const unsigned char * string)
{
    unsigned char char_pointer;                                 // Pointer at one string character.
    long int number;                                            // Temporary variable for calculating the result.
    
    number = 0;                                                 // Initialize variables.
    char_pointer=0;
    do                                                          // Process all charachters.
    {
        number *= 10;                                           // Multiply previous result by 10.
        number += string[char_pointer] - '0';                   // And add current digit.
        char_pointer++;
    }
    while (string[char_pointer] != 0);                          // Untill the null termination is found.

    return number;
}



// --------------------------------------------------------------------------
// This is the main program. After initialization the configuration is read
// from the EEPROM, a welcome screen is displayed (and sent over the RS232)
// and the program enters the main loop that polls all the inputs.
//
// First the main loops waits for a keypress (encoder / buttons / keypad) and
// while doing this it polls the serial port for incoming commands (serial
// characters are stored in rx_string and are processed when a [cr] is
// received).
//
// Then, with the encoder and the buttons the user can switch from
// VFO_A to VFO_B and vice versa, change VFO frequencies, VFO steps and save
// the current cdonfiguration into the EEPROM.
//
// Finally the keypad is cheked and if a keypress is detected the main loop
// enters a keypad loop that allows the handling of the pressed keys.
// --------------------------------------------------------------------------
void main(void)
{



    // Variables.

    unsigned char encoder_tmp;                                  // Bit 0 and 1 are the actual status of encoder channels A and B, bit 2 and 3 are previous status of A and B.
    unsigned char portb_tmp;                                    // Temporary register to see if PORTB has changed.

    unsigned char typed_string[10];                             // Nine character string used to store digits entered via the keypad.
    unsigned char typed_key, num_typed_keys, dot_pressed;       // Other variables used by the keypad: the last typed key, the total number of keys and if the decimal point has been used or not.

    unsigned char rx_string[13];                                // Twelve character string used for RS232 reception.
    unsigned char num_rx_chars;                                 // Number of received characters over the RS232.
    signed int rx_data;                                         // Temporary variable used to store received RS232 data.

    signed long int temp_freq;                                  // Used to convert ASCII frequency into a number (before assigning it to a VFO).
    unsigned char i;                                            // General purpose dummy variable.


    // Initialization.

    IO_init();                                                  // Initialize I/O lines.
    EEPROM_load_config();                                       // Initialize global variables.
    LCD_init();                                                 // Initialize LCD display.
    RS232_init();                                               // Initialize RS232 port
    DDS_init();                                                 // Initialize DDS.

    rx_string[0] = 0;
    num_rx_chars = 0;



    // Welcome screen.

    LCD_clear();
    LCD_home();
    LCD_line_1();                                               // Display welcome screen.
    LCD_write_string(WELCOME_STRING_1);
    LCD_line_2();
    LCD_write_string(WELCOME_STRING_2);

    RS232_send_string("\r\n");                                  // Display welocme screen over the RS232 terminal.
    RS232_send_string(WELCOME_STRING_1);
    RS232_send_string("\r\n");
    RS232_send_string(WELCOME_STRING_2);
    RS232_send_string("\r\n");

    delay(welcome_time);                                        // Wait 1s



    // Main screen.

    LCD_clear();                                                // Clear the screen.

    display_vfo(VFO_A);                                         // Prepare the main screen with frequencies of both VFOs.
    display_vfo(VFO_B);

    if (VFO_active == VFO_A)                                    // Set DDS to the frequency of the active VFO.
        DDS_set_freq(VFO_A_freq);
    else
        DDS_set_freq(VFO_B_freq);

    VFO_edit_mode = VFO_edit_freq;                              // By default the encoder will modify the freqeuncy.
    encoder_tmp = (PORTB >> 3) & 0b00000011;                    // Initialize encoder register (to avoid wrong direction step when first used).



    // Main loop.

    while (1)                                                   // Endless loop tha polls PORTB waiting for user input.
    {

        portb_tmp = PORTB;
        while (portb_tmp == PORTB)                              // Loop here untill PORTB changes.
        {



            // RS232 handler.

            rx_data = RS232_receive_data();                     // Nothing is happening on PORTB: poll the RS232 port.
            if ((rx_data >= (signed int)'a') && (rx_data <= (signed int)'z'))   // Convert to upper case.
                rx_data -= 'a' - 'A';

            if ((rx_data != RS232_NONE) && (rx_data != RS232_ERR))  // Check if any data was received.
            {
                if (rx_data == '\r')                            // Check if a carriage return was received.
                {
                    if (num_rx_chars > 0)                       // Check command only if a string was received.
                    {

                        RS232_send_string(rx_string);           // Echo typed command.
                        RS232_send_string("\r\n");

                        if ((num_rx_chars > 2) && (rx_string[0] == 'F'))    // If first character is 'f' than it's a frequency set command.
                        {
                            if ((rx_string[1] == 'A') || (rx_string[1] == 'B')) // Make sure a valid VFO was specified.
                            {
                                for (i=2; i<num_rx_chars; i++)              // Make sure the rest of the string is composed by digits only.
                                    if ((rx_string[i] < '0') || (rx_string[i] > '9'))
                                        break;
                                if ((rx_string[i] == 0) && (num_rx_chars < 12)) // Check if we could reach the end of the string without finding any non-digit character (and check that the number is not too long).
                                {
                                    for (i=2; i<=num_rx_chars; i++) // Copy received number into typed_string (vaiable used by the keypad handler and "borrowed" here to save memory).
                                        typed_string[i-2] = rx_string[i];
                                    temp_freq = str_2_long_int(typed_string);   // Convert into long int.
                                    if ((temp_freq >= DDS_min_freq) && (temp_freq <= DDS_max_freq)) // Check if number is in the valid DDS range.
                                    {
                                        if (rx_string[1] == 'A')                // Set new frequency to the correct VFO.
                                        {
                                            VFO_A_freq = temp_freq;
                                            display_vfo(VFO_A);                 // Update screen.
                                            if (VFO_active == VFO_A)            // Set new DDS frequency if VFO is active.
                                                DDS_set_freq(VFO_A_freq);
                                        }
                                        else
                                        {
                                            VFO_B_freq = temp_freq;
                                            display_vfo(VFO_B);                 // Update screen.
                                            if (VFO_active == VFO_B)            // Set new DDS frequency if VFO is active.
                                                DDS_set_freq(VFO_B_freq);
                                        }
                                        RS232_send_string("Ok.\r\n");           // Acknolage the command.
                                    }
                                    else                        // Number is out of DDS range.
                                        RS232_send_string("Error: frequency outside of DDS.\r\n");
                                }
                                else                            // No: not a valid number.
                                    RS232_send_string("Error: invalid frequency (must be integer in Hz).\r\n");
                            }
                            else
                                RS232_send_string("Error: invalid VFO (must be 'A' or 'B').\r\n");
                        }

                        else if ((num_rx_chars == 2) && (rx_string[0] == 'S'))  // If first character is 's' than it's VFO set command.
                        {
                            if (rx_string[1] == 'A')            // Activate VFO_A.
                            {
                                VFO_active = VFO_A;             // Set VFO_A as active VFO.
                                DDS_set_freq(VFO_A_freq);       // And set corresponding frequency.
                                display_vfo(VFO_A);             // Update the screen (since active VFO changed both lines need to be updated).
                                display_vfo(VFO_B);
                                RS232_send_string("Ok.\r\n");
                            }
                            else if (rx_string[1] == 'B')       // Activate VFO_B.
                            {
                                VFO_active = VFO_B;             // Set VFO_B as active VFO.
                                DDS_set_freq(VFO_B_freq);       // And set corresponding frequency.
                                display_vfo(VFO_A);             // Update the screen (since active VFO changed both lines need to be updated).
                                display_vfo(VFO_B);
                                RS232_send_string("Ok.\r\n");
                            }
                            else                                // Uuups: VFO must be either 'a' or 'b'.
                                RS232_send_string("Error: invalid VFO (must be 'A' or 'B').\r\n");
                        }

                        else if ((num_rx_chars == 1) && (rx_string[0] == '?'))  // Display welocme screen over the RS232 terminal.
                        {
                            RS232_send_string("\r\n");
                            RS232_send_string(WELCOME_STRING_1);
                            RS232_send_string("\r\n");
                            RS232_send_string(WELCOME_STRING_2);
                            RS232_send_string("\r\n");
                        }

                        else                                    // If none of the above it's a wrong command.
                            RS232_send_string("Error: unknown command ('S'et or 'F'req).\r\n");

                        rx_string[0] = 0;                       // Prepare vaiable for next reception.
                        num_rx_chars = 0;
                    }
                }
                else
                {
                    if (num_rx_chars < 12)                      // Make sure receive string is not full.
                    {
                        rx_string[num_rx_chars] = rx_data;      // Add the received character to the string.
                        num_rx_chars++;                         // Increment counter.
                        rx_string[num_rx_chars] = 0;            // Keep string null terminated.
                    }
                }
            }            
        }



        // Encoder handler.

        encoder_tmp = encoder_tmp << 2;                         // Shift old encoder channels left by two bits.
        encoder_tmp &= 0b00001100;                              // Clear all other bits (not really needed, just to be sure).
        encoder_tmp |= (PORTB >> 3) & 0b00000011;               // Put new encoder channels as bit 0 and 1 (reading PORTB will reset interrupt mismatch condition).

        if ((encoder_tmp == 0b00000001) || (encoder_tmp == 0b00001110))         // The encoder incremented (only half possibilities are taken into account because the used encoder generates two electrical steps per mechanical step).
//        if ((encoder_tmp == 0b00000001) || (encoder_tmp == 0b00001110) || (encoder_tmp == 0b00000111) || (encoder_tmp == 0b00001000))   // The encoder incremented.
        {
            if (VFO_edit_mode == VFO_edit_freq)                 // Check if we are editing the frequency or the step.
            {
                if (VFO_active == VFO_A)                        // Edit frequency: edit the correct VFO.
                {
                    VFO_A_freq += VFO_steps[VFO_A_step];        // Increment the corresponding frequency step for each encoder step.
                    if (VFO_A_freq > DDS_max_freq)              // Make sure the maximum frequency of the DDS is not exceeded.
                        VFO_A_freq = DDS_max_freq;
                    DDS_set_freq(VFO_A_freq);                   // Set the new frequency in the DDS.
                    display_vfo(VFO_A);                         // Update the screen.
                }
                else
                {
                    VFO_B_freq += VFO_steps[VFO_B_step];        // Increment the corresponding frequency step for each encoder step.
                    if (VFO_B_freq > DDS_max_freq)              // Make sure the maximum frequency of the DDS is not exceeded.
                        VFO_B_freq = DDS_max_freq;
                    DDS_set_freq(VFO_B_freq);                   // Set the new frequency in the DDS.
                    display_vfo(VFO_B);                         // Update the screen.
                }
            }
            else
            {
                if (VFO_active == VFO_A)                        // Edit step: edit the correct VFO.
                {
                    VFO_A_step++;                               // Select next step.
                    if (VFO_A_step >= (VFO_num_steps - 1))      // Make sure the highest step is not exceeded.
                        VFO_A_step = VFO_num_steps - 1;
                    display_step(VFO_A);                        // Update the screen.
                }
                else
                {
                    VFO_B_step++;                               // Select next step.
                    if (VFO_B_step >= (VFO_num_steps - 1))      // Make sure the highest step is not exceeded.
                        VFO_B_step = VFO_num_steps - 1;
                    display_step(VFO_B);                        // Update the screen.
                }
            }
        }
        else if ((encoder_tmp == 0b00000010) || (encoder_tmp == 0b00001101))    // The encoder decremented (only half possibilities are taken into account because the used encoder generates two electrical steps per mechanical step).
//        else if ((encoder_tmp == 0b00000010) || (encoder_tmp == 0b00001101) || (encoder_tmp == 0b00001011) || (encoder_tmp == 0b00000100))  // The encoder decremented.
        {
            if (VFO_edit_mode == VFO_edit_freq)                 // Check if we are editing the frequency or the step.
            {
                if (VFO_active == VFO_A)                        // Edit frequency: edit the correct VFO.
                {
                    VFO_A_freq -= VFO_steps[VFO_A_step];        // Decrement the corresponding frequency step for each encoder step.
                    if (VFO_A_freq < DDS_min_freq)              // Make sure the minimum frequency of the DDS is not exceeded.
                        VFO_A_freq = DDS_min_freq;
                    DDS_set_freq(VFO_A_freq);                   // Set the new frequency in the DDS.
                    display_vfo(VFO_A);                         // Update the screen.
                }
                else
                {
                    VFO_B_freq -= VFO_steps[VFO_B_step];        // Decrement the corresponding frequency step for each encoder step.
                    if (VFO_B_freq < DDS_min_freq)              // Make sure the minimum frequency of the DDS is not exceeded.
                        VFO_B_freq = DDS_min_freq;
                    DDS_set_freq(VFO_B_freq);                   // Set the new frequency in the DDS.
                    display_vfo(VFO_B);                         // Update the screen.
                }
            }
            else
            {
                if (VFO_active == VFO_A)                        // Edit step: edit the correct VFO.
                {
                    VFO_A_step--;                               // Select previous step.
                    if (VFO_A_step < 0)                         // Make sure the lowest step is not exceeded.
                        VFO_A_step = 0;
                    display_step(VFO_A);                        // Update the screen.
                }
                else
                {
                    VFO_B_step--;                               // Select previous step.
                    if (VFO_B_step < 0)                         // Make sure the lowest step is not exceeded.
                        VFO_B_step = 0;
                    display_step(VFO_B);                        // Update the screen.
                }
            }
        }
        else
        {



            // Button handler.

            if ((PORTB & 0b11100000) != 0b11100000)             // Check buttons.
            {

                if (ENC_BTN == 0)                               // Check if the encoder button was pressed (used to toggle between frequency edit mode and step edit mode).
                {
                    delay(debounce_time);                       // Yes: debounce.
                    while (ENC_BTN == 0) ;                      // Wait untill button is released.
                    delay(debounce_time);                       // Debounce again.
                    if (VFO_edit_mode == VFO_edit_freq)         // Toggle between frequency edit mode and step edit mode.
                    {
                        VFO_edit_mode = VFO_edit_step;          // Toggle mode.
                        if (VFO_active == VFO_A)                // Prepare display for step editing.
                            display_step(VFO_A);
                        else
                            display_step(VFO_B);
                    }
                    else
                    {
                        VFO_edit_mode = VFO_edit_freq;          // Toggle mode.
                        if (VFO_active == VFO_A)                // Prepare display for frequency editing.
                            display_vfo(VFO_A);
                        else
                            display_vfo(VFO_B);
                    }
                }

                if (BTN1 == 0)                                  // Check if button 1 was presssed (used to select VFO_A and VFO_B, but also to exit step edit mode).
                {
                    delay(debounce_time);                       // Yes: debounce.
                    while (BTN1 == 0) ;                         // Wait untill button is released.
                    delay(debounce_time);                       // Debounce again.
                    if (VFO_edit_mode == VFO_edit_freq)         // Check edit mode: frequency or step.
                    {
                        if (VFO_active == VFO_A)                // In frequency edit mode this button toggles between VFO_A and VFO_B.
                        {
                            VFO_active = VFO_B;                 // Set active VFO.
                            DDS_set_freq(VFO_B_freq);           // And set corresponding frequency.
                        }
                        else
                        {
                            VFO_active = VFO_A;                 // Set active VFO.
                            DDS_set_freq(VFO_B_freq);           // And set corresponding frequency.
                        }
                        display_vfo(VFO_A);                     // Update the screen (since active VFO changed both lines need to be updated).
                        display_vfo(VFO_B);
                    }
                    else
                    {
                        VFO_edit_mode = VFO_edit_freq;          // In step edit mode this button switches back into frequency edit mode.
                        if (VFO_active == VFO_A)                // And prepare display for frequency editing.
                            display_vfo(VFO_A);
                        else
                            display_vfo(VFO_B);
                    }
                }
                if (BTN2 == 0)                                  // Check if button 2 was presssed (used to save configuration, but also to exit step edit mode).
                {
                    delay(debounce_time);                       // Yes: debounce.
                    while (BTN2 == 0) ;                         // Wait untill button is released.
                    delay(debounce_time);                       // Debounce again.
                    if (VFO_edit_mode == VFO_edit_freq)         // Check edit mode: frequency or step.
                    {
                        EEPROM_save_config();                   // In frequency edit mode this button saves the configuration into the EEPROM.
                        LCD_clear();                            // Tell the user that the configuration was saved.
                        LCD_line_1();
                        LCD_write_string("Configuration   ");
                        LCD_line_2();
                        LCD_write_string("saved.          ");
                        delay(message_time);                    // Wait 1.0 s
                        LCD_clear();                            // Restore main screen.
                        display_vfo(VFO_A);
                        display_vfo(VFO_B);
                    }
                    else
                    {
                        VFO_edit_mode = VFO_edit_freq;          // In step edit mode this button switches back into frequency edit mode.
                        if (VFO_active == VFO_A)                // And prepare display for frequency editing.
                            display_vfo(VFO_A);
                        else
                            display_vfo(VFO_B);
                    }
                }
            }



            // Keypad handler.

            else if ((PORTB & 0b00000111) != 0b00000111)        // Check keypad.
            {
                if (VFO_edit_mode == VFO_edit_freq)             // Only typing of the VFO frequency is allowed.
                {
                    if (VFO_active == VFO_A)                    // Prepare display for keypad edit.
                        LCD_line_1();
                    else
                        LCD_line_2();
                    LCD_write_string("                ");
                    if (VFO_active == VFO_A)
                        LCD_line_1();
                    else
                        LCD_line_2();
                    LCD_on_cursor();                            // Switch cursor on.

                    LCD_write_string("> ");                     // Display prompt.

                    typed_key = read_keypad();                  // If the '#' was pressed wait untill keypad is released.
                    if (typed_key == '#')
                    {
                        while (read_keypad() != 0);
                        delay(debounce_time);                   // Debounce.
                    }

                    for (i=0; i<9; i++)                         // Initialize typed string to 000000000.
                        typed_string[i] = '0';
                    typed_string[9] = 0;
                    num_typed_keys = 0;
                    dot_pressed = 0;

                    while (1)                                   // Keypad edit loop (exits with break).
                    {
                        do                                      // Wait untill the keypad is pressed.
                            typed_key = read_keypad();
                        while (typed_key == 0);
                        delay(debounce_time);                   // Debounce.
                        while (typed_key == read_keypad());     // Wait untill keypad is released.
                        delay(debounce_time);                   // Debounce again.

                        if (typed_key == '#')                   // The '#' key (enter) will exit the loop.
                            break;

                        if (typed_key == '*')
                        {
                            if (dot_pressed == 0)               // Was the decimal point already pressed?
                            {
                                dot_pressed = 1;                // No: it is pressed now.
                                if (num_typed_keys <= 3)        // If more than 3 digits have already been pressed no dot is possible.
                                {
                                    for (i=0; i<3; i++)         // Shift all digits by 6 (dot means MHz).
                                        typed_string[i] = typed_string[i+6];
                                    for (i=3; i<9; i++)
                                        typed_string[i] = '0';  // Remove original digits.
                                    if (num_typed_keys == 0)    // If the first key is '*' display "0.".
                                        LCD_write_data('0');
                                    num_typed_keys = 3;         // Pretend 3 key have been pressed.
                                    LCD_write_data('.');        // Display a '.'.
                                }
                                else
                                    break;                      // Invalid point: exit the loop (escape).
                            }
                            else
                                break;                          // Yes: exit the loop (escape).
                        }
                        else
                        {
                            if (dot_pressed == 0)               // Was the decimal point already pressed?
                            {
                                for (i=0; i<8; i++)             // No: hift all typed keys left and insert the new key.
                                    typed_string[i] = typed_string[i+1];
                                typed_string[8] = typed_key;
                                num_typed_keys++;
                            }
                            else
                            {
                                typed_string[num_typed_keys] = typed_key;   // Yes: set character at correct position.
                                num_typed_keys++;
                            }
                            LCD_write_data(typed_key);
                        }
                        if (num_typed_keys >= 9)                // When all possible digits (9) have been typed, do as if enter ('#') was pressed.
                        {
                            typed_key = '#';
                            break;
                        }
                    }

                    if (typed_key != '*')                       // Convert to frequency only if escape ('*') was not pressed.
                    {
                        temp_freq = str_2_long_int(typed_string);   // Convert typed string into a number.
                        if ((temp_freq >= DDS_min_freq) && (temp_freq <= DDS_max_freq)) // Check if number is in the valid DDS range.
                            if (VFO_active == VFO_A)            // Set new frequency to the correct VFO.
                            {
                                VFO_A_freq = temp_freq;
                                DDS_set_freq(VFO_A_freq);
                            }
                            else
                            {
                                VFO_B_freq = temp_freq;
                                DDS_set_freq(VFO_B_freq);
                            }
                    }

                    LCD_on_nocursor();                          // Switch cursor off.
                    if (VFO_active == VFO_A)                    // And restore VFO display.
                        display_vfo(VFO_A);
                    else
                        display_vfo(VFO_B);
                    encoder_tmp = (PORTB >> 3) & 0b00000011;    // Initialize encoder register again (to avoid wrong direction step when first used).
                }
            }
        }



    }
}


